@@ -88,7 +88,7 @@ typedef struct {
 
 #if defined(__STDC__) || defined(_MSDOS) || defined(_WIN32)
 extern void krb5_MD4Init(krb5_MD4_CTX FAR *);
-extern void krb5_MD4Update(krb5_MD4_CTX FAR *, unsigned char FAR *, unsigned int);
+extern void krb5_MD4Update(krb5_MD4_CTX FAR *, const unsigned char FAR *, unsigned int);
 extern void krb5_MD4Final(krb5_MD4_CTX FAR *);
 #else
 void krb5_MD4Init ();
