@@ -27,12 +27,12 @@
  * Find default credential cache
  */
 
-#include "k5-int.h"
-
 #ifdef USE_LOGIN_LIBRARY
-#include <KerberosLoginPrivate/KerberosLoginPrivate.h>
+#include <Kerberos/KerberosLoginPrivate.h>
 #endif
 
+#include "k5-int.h"
+
 KRB5_DLLIMP krb5_error_code KRB5_CALLCONV
 krb5_cc_default(context, ccache)
    krb5_context context;
@@ -98,7 +98,8 @@ krb5int_cc_default(context, ccache)
 		}
 		
 		/* Try to make sure a krb5 tgt is in the cache */
-		err = __KLInternalAcquireTicketsForCache (desiredPrincipal, krb5_cc_default_name(context), 
+		err = __KLInternalAcquireInitialTicketsForCache (desiredPrincipal, NULL, 
+                                                    krb5_cc_default_name (context), 
 													kerberosVersion_V5, nil, &outCacheName);
 		if (err == klNoErr) {
 			/* This function tries to get tickets and put them in the specified 
