@@ -173,7 +173,7 @@ init_common (context, secure)
 			    "kdc_default_options", 0,
 			    KDC_OPT_RENEWABLE_OK, &tmp);
 	ctx->kdc_default_options = KDC_OPT_RENEWABLE_OK;
-#ifdef macintosh
+#if TARGET_OS_MAC
 #define DEFAULT_KDC_TIMESYNC 1
 #else
 #define DEFAULT_KDC_TIMESYNC 0
@@ -191,7 +191,7 @@ init_common (context, secure)
 	 * Note: DCE 1.0.3a only supports a cache type of 1
 	 * 	DCE 1.1 supports a cache type of 2.
 	 */
-#ifdef macintosh
+#if TARGET_OS_MAC
 #define DEFAULT_CCACHE_TYPE 4
 #else
 #define DEFAULT_CCACHE_TYPE 3
@@ -374,8 +374,8 @@ krb5_get_default_in_tkt_ktypes(context, ktypes)
 				  context->in_tkt_ktypes));
 }
 
-krb5_error_code
-krb5_set_default_tgs_ktypes(context, ktypes)
+krb5_error_code KRB5_CALLCONV
+krb5_set_default_tgs_enctypes (context, ktypes)
 	krb5_context context;
 	const krb5_enctype *ktypes;
 {
@@ -406,6 +406,13 @@ krb5_set_default_tgs_ktypes(context, ktypes)
     return 0;
 }
 
+krb5_error_code krb5_set_default_tgs_ktypes
+(krb5_context context, const krb5_enctype *etypes)
+{
+  return (krb5_set_default_tgs_enctypes (context, etypes));
+}
+
+
 void
 KRB5_CALLCONV
 krb5_free_ktypes (context, val)
